@@ -15,7 +15,7 @@
  */
 
 #include "mongoc-config.h"
-#include <bson.h>
+#include <bson/bson.h>
 
 #ifdef MONGOC_ENABLE_CRYPTO_LIBCRYPTO
 #include "mongoc-crypto-openssl-private.h"
@@ -38,14 +38,14 @@ mongoc_crypto_openssl_hmac_sha1 (mongoc_crypto_t *crypto,
    HMAC (EVP_sha1 (), key, key_len, data, data_len, hmac_out, NULL);
 }
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L || (defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER < 0x2070000fL)
-EVP_MD_CTX *
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || (defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER < 0x20700000L)
+static EVP_MD_CTX *
 EVP_MD_CTX_new (void)
 {
    return bson_malloc0 (sizeof (EVP_MD_CTX));
 }
 
-void
+static void
 EVP_MD_CTX_free (EVP_MD_CTX *ctx)
 {
    EVP_MD_CTX_cleanup (ctx);
